/* {{{ Includes */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/wait.h>

#include <tcpa.h>

#define CONFIG_ENFORCER_TCPA 1
#include "../kernel/enforcer_common.h"

#define SECRET_SIZE 256

/* }}} */

/* {{{ Globals */

/* Read from external config files */
char *pass1;
char *pass2;

char *blob_file = ENFORCER_TCPA_BLOB_FILE;
char *alg = "aes";
char *loop_device = "/dev/loop7";
char *losetup_program = ENFORCER_FILE_BASE "/losetup";
char *mount_program = ENFORCER_FILE_BASE "/mount";
char *umount_program = ENFORCER_FILE_BASE "/umount";

char *loopback_file = ENFORCER_FILE_BASE "/loop.data";
char *mount_point = ENFORCER_FILE_BASE "/loop";

int lock = 1;					/* tpm-lock */

char *PROGRAM;

/* }}} */

/* {{{ int parse_configs(void) */

typedef struct {
   const char *name;
   const char *type;
   void *variable;
} entry_t;

static entry_t entries[] = {
   {"tcpa-blob-file", "char", &blob_file},
   {"encryption-algorithm", "char", &alg},
   {"loopback-device", "char", &loop_device},
   {"losetup-program", "char", &losetup_program},
   {"mount-program", "char", &mount_program},
   {"umount-program", "char", &umount_program},
   {"loopback-file", "char", &loopback_file},
   {"mount-point", "char", &mount_point},
   {"tpm-lock", "int", &lock},
   {0, 0, 0}
};

int process_entry(char *line) {
   char *value;
   entry_t *entry;
   int result = 0;

   for(entry = &entries[0]; entry->name != NULL; entry++) {
	  if(strncasecmp(entry->name, line, strlen(entry->name)) == 0) {
		 value = line + strlen(entry->name);
		 if(*value != '=') {
			fprintf(stderr, "%s: error, expecting an equal sign, got: %s\n",
					PROGRAM, value);
			result = 1;
			goto out;
		 }
		 value++;					/* bypass the '=' */

		 if(strcasecmp("char", entry->type) == 0) {
			char **cvariable = (char **)entry->variable;
			int len = strlen(value);
			*cvariable = (char *)malloc(len + 1);
			if(!malloc) {
			   result = errno;
			   fprintf(stderr, "%s: unable to allocate %d bytes of memory: %s\n",
					   PROGRAM, len, strerror(result));
			   goto out;
			}
			memset(*cvariable, 0, len + 1);
			strncpy(*cvariable, value, len);
		 } else if(strcasecmp("int", entry->type) == 0) {
			int *ivariable = (int *)entry->variable;
			if(strncasecmp("1", value, 1) == 0 ||
			   strncasecmp("on", value, 2) == 0 ||
			   strncasecmp("yes", value, 3) == 0 ||
			   strncasecmp("true", value, 4) == 0 ||
			   strncasecmp("enable", value, 6) == 0
			   ) {
											   
			   *ivariable = 1;
			} else {
			   *ivariable = 0;
			}
		 }
	  }
   }

 out:
   return result;
}

int parse_configs(void) {
   int result = 0;
   FILE *fd;
   char buffer[4096];
   char *tmp, *str;
   int len;
   
   fd = fopen(ENFORCER_TCPA_PASSWORD_FILE, "r");
   if(!fd) {
	  fprintf(stderr, "%s: unable to open '%s' for reading.\n",
			  PROGRAM, ENFORCER_TCPA_PASSWORD_FILE);
	  result = 1;
	  goto out;
   }

   memset(buffer, 0, sizeof(buffer));
   result = fread(buffer, 4096, 1, fd);
   if(fd <= 0) {
	  result = errno;
	  fprintf(stderr, "%s: error reading from '%s': %s.\n",
			  PROGRAM, ENFORCER_TCPA_PASSWORD_FILE,
			  strerror(errno));
	  goto out;
   }

   tmp = buffer;
   /* Read 1st password */
   str = strsep(&tmp, "\n");
   if(!str) {
	  result = 1;
	  fprintf(stderr, "%s: error reading password #1 from '%s'.\n",
			  PROGRAM, ENFORCER_TCPA_PASSWORD_FILE);
	  goto out;
   }

   len = strlen(str);
   pass1 = (char *)malloc(len + 1);
   if(!pass1) {
	  result = errno;
	  fprintf(stderr, "%s: unable to allocate %d bytes of memory: %s\n",
			  PROGRAM, len, strerror(result));
	  goto out;
   }
   
   memset(pass1, 0, len + 1);
   strncpy(pass1, str, len);

   /* Read 2nd password */
   str = strsep(&tmp, "\n");
   if(!str) {
	  result = 1;
	  fprintf(stderr, "%s: error reading password #2 from '%s'.\n",
			  PROGRAM, ENFORCER_TCPA_PASSWORD_FILE);
	  goto out;
   }

   len = strlen(str);
   pass2 = (char *)malloc(len + 1);
   if(!pass2) {
	  result = errno;
	  fprintf(stderr, "%s: unable to allocate %d bytes of memory: %s\n",
			  PROGRAM, len, strerror(result));
	  goto out;
   }
   
   memset(pass2, 0, len + 1);
   strncpy(pass2, str, len);

   fclose(fd);


   /* Read 2nd config file */
   fd = fopen(ENFORCER_HELPER_CONFIG_FILE, "r");
   if(!fd) {
	  fprintf(stderr, "%s: unable to open '%s' for reading.\n",
			  PROGRAM, ENFORCER_HELPER_CONFIG_FILE);
	  result = 1;
	  goto out;
   }

   memset(buffer, 0, sizeof(buffer));
   result = fread(buffer, 4096, 1, fd);
   if(fd <= 0) {
	  result = errno;
	  fprintf(stderr, "%s: error reading from '%s': %s.\n",
			  PROGRAM, ENFORCER_HELPER_CONFIG_FILE,
			  strerror(errno));
	  goto out;
   }

   tmp = buffer;
   while((str = strsep(&tmp, "\n")) != NULL) {
	  result = process_entry(str);
	  if(result)
		 goto out;
   }

#if 0
   {
	  entry_t *entry;
	  printf("---------- helper.conf ----------\n");
	  for(entry = &entries[0]; entry->name != NULL; entry++) {
		 if(strcasecmp("char", entry->type) == 0) {
			char **cvariable = (char **)entry->variable;
			printf("%s=%s\n", entry->name, *cvariable);
		 } else if(strcasecmp("int", entry->type) == 0) {
			int *ivariable = (int *)entry->variable;
			printf("%s=%d\n", entry->name, *ivariable);
		 }
	  }
   }
#endif

   
 out:
   if(fd)
	  fclose(fd);

   return(result);
}

/* }}} */

/* {{{ char *tpm_unseal(int *datalen) */

char *tpm_unseal(int *datalen) {
   /* Based on IBM's tpm/examples/unsealfile.c */
   
   unsigned char passhash1[20];    /* hash of parent key password */
   unsigned char passhash2[20];    /* hash of data password */
   uint32_t ret;
   struct stat sbuf;

   unsigned char blob[4096];	/* sealed blob */
   unsigned int bloblen;		/* blob length */
   FILE *infile;
   unsigned char *databuff;

   databuff = (unsigned char *)malloc(SECRET_SIZE);
   if(!databuff) {
	  fprintf(stderr, "%s: error allocating %d bytes of memory: %s\n",
			  PROGRAM, SECRET_SIZE, strerror(errno));
	  return(NULL);
   }

   if(mlock(databuff, SECRET_SIZE) != 0) {
	  ret = errno;
	  fprintf(stderr, "%s: unable to lock critical pages in memory: %s\n",
			  PROGRAM, strerror(errno));
	  goto out;
   }

   memset(databuff, 0, SECRET_SIZE);
   
   infile = fopen(blob_file, "r");
   if (infile == NULL) {
	  ret = errno;
	  fprintf(stderr, "%s: unable to open input file '%s': %s\n",
			  PROGRAM, blob_file, strerror(errno));
	  goto out_unlock;
   }

   stat(blob_file, &sbuf);
   bloblen = (int) sbuf.st_size;
   ret = fread(blob, 1, bloblen, infile);
   if (ret != bloblen) {
	  ret = errno;
	  fprintf(stderr, "%s: I/O Error while reading input file '%s': %s\n",
			  PROGRAM, blob_file, strerror(errno));
	  goto out_unlock;
   }
   fclose(infile);

   sha1(pass1, strlen(pass1), passhash1);
   sha1(pass2, strlen(pass2), passhash2);

   /* Finally, do the unseal */
   ret = TPM_Unseal(0x40000000,	/* KEY Entity Value */
					passhash1,          /* Key Password */
					passhash2,          /* blob password */
					blob, bloblen,      /* encrypted blob, blob length */
					databuff, datalen, /* buffer for decrypted data, len */
					NULL);

   if(ret != 0) {
	  fprintf(stderr, "%s: Error %d from TPM_Unseal.\n",
			  PROGRAM, ret);
	  ret = 1;
	  goto out_unlock;
   }
   
   return(databuff);

 out_unlock:
   if(munlockall() != 0) {
	  fprintf(stderr, "%s: unable to unlock locked memory: %s\n",
			  PROGRAM, strerror(errno));
   }
   
 out:
   free(databuff);
   
   return (NULL);
}

/* }}} */

/* {{{ int tpm_lock(void) */

int tpm_lock(void) {
   char hash_and_extend_fmt[] = "00 C1 T 00 00 00 14 L %";
   unsigned char randomish[20];
   FILE *fd;
   int result = 0;
   char buffer[TCPA_MAX_BUFF_SIZE];
   
   fd = fopen("/dev/urandom", "r");
   if(fd) {
	  result = fread(randomish, 20, 1, fd);
	  fclose(fd);
   }

   if(!fd || result != 20) {
	  strncpy(randomish, "thooX8ooleemiengawit", 20);
   }

   memset(buffer, 0, TCPA_MAX_BUFF_SIZE);
   
   result = buildbuff(hash_and_extend_fmt, buffer,
					  ENFORCER_PUBLIC_KEY_PCR, 20, randomish);

   if(result <= 0) {
	  result = 1;
	  goto out;
   }

   result = TPM_Transmit(buffer, NULL, NULL);

 out:
   return result;
}

/* }}} */

/* {{{ int stop(void) */

int stop(void) {
   int filedes[2];
   int result = 0;
   pid_t child_pid;

   child_pid = fork();
   
   if(child_pid > 0) {
	  /* In parent */
	  pid_t wait_pid;

	  wait_pid = waitpid(child_pid, &result, 0);

	  if(!(WIFEXITED(result) && WEXITSTATUS(result) == 0)) {
		 /* Child exited abnormally */
		 fprintf(stderr, "%s: umount child exited non-zero (%d).\n",
				 PROGRAM, WEXITSTATUS(result));
		 result = 1;
		 goto out;	 
	  }
   } else if (child_pid == 0) {
	  /* In child */
	  char *envp[] = { NULL };
	  
	  char *argv[] = { "umount",
					   "-d", mount_point,
					   NULL };

	  /* sudo umount /home/omen/d/enforcer/loop */
	  result = execve(umount_program, argv, envp);
	  fprintf(stderr, "%s: (child) execve error, should never reach here: %s\n",
			  PROGRAM, result, strerror(errno));
	  exit(1);
   } else {
	  /* Error */
	  result = errno;
	  fprintf(stderr, "%s: error forking for umount: %s\n",
			  PROGRAM, strerror(result));
	  goto out;
   }
   

 out:
   return result;
}

/* }}} */

/* {{{ int start(void) */

int start(void) {
   int filedes[2];
   int result = 0;
   pid_t child_pid;
   int all_clear = 1;
   int key_len;
   char *key;
   
   result = pipe(filedes);
   if(result < 0) {
	  result = errno;
	  fprintf(stderr, "%s: error setting up pipes for mount: %s\n",
			  PROGRAM, strerror(result));
	  goto out;
   }

   /* Read from TCPA */
   key = tpm_unseal(&key_len);
   
   if(key == NULL) {
	  result = 1;
	  fprintf(stderr, "%s: error retrieving key from TPM.\n",
			  PROGRAM);
	  all_clear = 0;
   }
   
   if(all_clear && lock) {
	  result = tpm_lock();
	  if(result != 0) {
		 
		 /* Destroy secret key in memory */
		 memset(key, 0, key_len);
		 if(munlockall() != 0) {
			fprintf(stderr, "%s: unable to unlock locked memory: %s\n",
					PROGRAM, strerror(errno));
		 }

		 fprintf(stderr, "%s: error %d locking the TPM.\n",
				 PROGRAM, result);
		 result = 1;
		 all_clear = 0;
	  }
   }

   if(all_clear) {
	  child_pid = fork();
   
	  if(child_pid > 0) {
		 /* In parent */
		 pid_t wait_pid;
	  
		 close(filedes[0]);		/* read pipe */

		 result = write(filedes[1], key, key_len);

		 /* Destroy secret key in memory */
		 memset(key, 0, key_len);
		 if(munlockall() != 0) {
			fprintf(stderr, "%s: unable to unlock locked memory: %s\n",
					PROGRAM, strerror(errno));
		 }

		 if(result != key_len) {
			result = errno;
			fprintf(stderr, "%s: error writing to losetup pipe: %s\n",
					PROGRAM, strerror(result));
			all_clear = 0;
		 }

		 if(close(filedes[1]) < 0) {
			result = errno;
			fprintf(stderr, "%s: error closing losetup pipe: %s\n",
					PROGRAM, strerror(result));
			all_clear = 0;
		 }
	  
		 wait_pid = waitpid(child_pid, &result, 0);

		 if(!(WIFEXITED(result) && WEXITSTATUS(result) == 0)) {
			/* Child exited abnormally */
			fprintf(stderr, "%s: losetup child exited non-zero (%d).\n",
					PROGRAM, WEXITSTATUS(result));
			result = 1;
			all_clear = 0;
		 }

	  } else if (child_pid == 0) {
		 /* In child */
		 char *envp[] = { NULL };
		 char pfd[2];
		 char *argv[] = { "losetup",
						  "-p", pfd, "-e", alg, loop_device, loopback_file,
						  NULL };
		 
		 memset(key, 0, SECRET_SIZE);
	  
		 snprintf(pfd, 2, "%1d", filedes[0]);

		 close(filedes[1]);		/* write pipe */
	  
		 /* losetup -p 0 -e aes /dev/loop0 ./crypto.loop */
		 result = execve(losetup_program, argv, envp);
		 fprintf(stderr, "%s: (child) execve error, should never reach here: %s\n",
				 PROGRAM, result, strerror(errno));
		 exit(1);
	  } else {
		 /* Error */
		 result = errno;
		 fprintf(stderr, "%s: error forking for losetup: %s\n",
				 PROGRAM, strerror(result));
		 all_clear = 0;
	  }
   }


   if(all_clear) {
	  /* Parent code, after first fork */
	  child_pid = fork();
   
	  if(child_pid > 0) {
		 /* In parent */
		 pid_t wait_pid;

		 wait_pid = waitpid(child_pid, &result, 0);

		 if(!(WIFEXITED(result) && WEXITSTATUS(result) == 0)) {
			/* Child exited abnormally */
			fprintf(stderr, "%s: mount child exited non-zero (%d).\n",
					PROGRAM, WEXITSTATUS(result));
			result = 1;
			goto out;	 
		 }
	  } else if (child_pid == 0) {
		 /* In child */
		 char *envp[] = { NULL };

		 char *argv[] = { "mount",
						  loop_device, mount_point,
						  NULL };

		 /* sudo mount /dev/loop7 /home/omen/d/enforcer/loop */
		 result = execve(mount_program, argv, envp);
		 fprintf(stderr, "%s: (child) execve error, should never reach here: %s\n",
				 PROGRAM, result, strerror(errno));
		 exit(1);
	  } else {
		 /* Error */
		 result = errno;
		 fprintf(stderr, "%s: error forking for mount: %s\n",
				 PROGRAM, strerror(result));
		 goto out;
	  }
   }  

 out:
   if(key)
	  free(key);
   
   return result;
}

/* }}} */

/* {{{ int main(int argc, char *argv[]) */

int main(int argc, char *argv[]) {
   int result = 0;

   PROGRAM = strrchr(argv[0], '/');
   PROGRAM = PROGRAM ? PROGRAM+1 : argv[0];

   if(argc != 2) {
	  fprintf(stderr, "%s: usage\n", PROGRAM);
	  fprintf(stderr, "%s (start|stop|force-stop|tpm-lock)\n");
	  exit(1);
   }

   result = parse_configs();
   if(result != 0)
	  goto out;
   
   if(strncmp("start", argv[1], 5) == 0) {
	  result = start();
   } else if(strncmp("stop", argv[1], 4) == 0) {
	  result = stop();
   } else if(strncmp("force-stop", argv[1], 10) == 0) {
	  /* Called from inside the kernel */
	  result = stop();
	  if(result != 0) {
		 /* Work around call_usermodehelper bug (as of 2.6-test9) */
		 /* sync, shutdown */
	  }
   } else if(strncmp("tpm-lock", argv[1], 8) == 0) {
	  result = tpm_lock();
   } else {
	  result = 1;
   }

 out:
  return(result);
}

/* }}} */
