/* verify.c */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "bigdigits.h"

/* MOD_SIZE * 32 = #bits */ 
#define MOD_SIZE 64

int genRsaPrime( DIGIT_T p[], unsigned int ndigits );
void dumpKey( DIGIT_T k[], char *fname, unsigned int len );


int main( int argc, char *argv[] )
{

    DIGIT_T n[MOD_SIZE], e[MOD_SIZE], d[MOD_SIZE];	
    DIGIT_T p[MOD_SIZE], q[MOD_SIZE];
    DIGIT_T L[2*MOD_SIZE];
    DIGIT_T g[MOD_SIZE], t[MOD_SIZE];
    DIGIT_T FermatNumber[] = { 0x3, 0x5, 0x11, 0x101, 0x10001 };

    unsigned f = 4;
    unsigned int ndigits;
    int status;

    char *pub_file;
    char *prv_file;
	
    /* Command line args */
    if (argc < 3)
    {
	printf( "Usage: genkey <public_key_file> <private_key_file>\n" );
	exit( 1 );
    }
    else
    {
	pub_file = argv[1];
	prv_file = argv[2];
    }

    /* Set e */
    mpSetZero(e, MOD_SIZE);
    if (f > 4) f = 4;
    e[0] = FermatNumber[f];
	
    /* Generate random prime p, ~ half length of modulus */
    mpSetZero(p, MOD_SIZE);
    ndigits = MOD_SIZE / 2;
    do
    {
	status = genRsaPrime(p, ndigits);
	if (status) return -1;

	/* Make sure gcd(p-1, e) = 1 */
	mpShortSub(L, p, 1, ndigits);
	mpGcd(g, L, e, ndigits);

#ifdef VERBOSE
	printf("g="); mpPrintNL(g, ndigits);
#endif
    } while (mpShortCmp(g, 1, ndigits) != 0);

	
#ifdef VERBOSE
    printf("Final p="); mpPrintTrimNL(p, ndigits);
#endif

    /* And random prime q for balance of modulus length */
    mpSetZero(q, MOD_SIZE);
    ndigits = MOD_SIZE - ndigits;
    do
    {
	status = genRsaPrime(q, ndigits);
	if (status) return -1;
	
	/* Make sure gcd(q-1, e) = 1 */
	mpShortSub(L, q, 1, ndigits);
	mpGcd(g, L, e, ndigits);

#ifdef VERBOSE
	printf("g="); mpPrintTrimNL(g, ndigits);
#endif

    } while (mpShortCmp(g, 1, ndigits) != 0);


#ifdef VERBOSE
    printf("Final q="); mpPrintNL(q, ndigits);
    printf("Final p="); mpPrintTrimNL(p, MOD_SIZE);
#endif
    
    /*	Calculate modulus, n = pq 
	using L as a double-length temp */
    mpMultiply(L, p, q, MOD_SIZE);	/* L is double-length */
    mpSetEqual(n, L, MOD_SIZE);		/* but n is only single */
    
    printf("\nModulus:\n"); mpPrintNL(n, MOD_SIZE);
    
    /* L = (p-1)*(q-1) */
    mpShortSub(g, p, 1, MOD_SIZE);
    mpShortSub(t, q, 1, MOD_SIZE);
    mpMultiply(L, g, t, MOD_SIZE);

#ifdef VERBOSE
    printf("L="); mpPrintNL(L, MOD_SIZE);
#endif

    /* Calculate private key, d = e^-1 Mod L */
    mpModInv(d, e, L, MOD_SIZE);
    
    printf("\nPrivate Exponent:\n"); mpPrintNL(d, MOD_SIZE);
    printf("\nPublic Exponent:\n"); mpPrintTrimNL(e, MOD_SIZE);
    
    dumpKey( n, pub_file, MOD_SIZE );
    dumpKey( d, prv_file, MOD_SIZE );

    /* Clear up */
    mpSetZero(p, MOD_SIZE);
    mpSetZero(q, MOD_SIZE);
    mpSetZero(g, MOD_SIZE);
    mpSetZero(t, MOD_SIZE);
    mpSetZero(L, 2*MOD_SIZE);
    
    return 0;
}


/*=======================================================
 * Dump the key to fname.
 ======================================================*/
void dumpKey( DIGIT_T k[], char *fname, unsigned int len )
{
    FILE *fp;
    int i = 0;

    fp = fopen( fname, "w" );
    if (fp == NULL)
    {
	printf("Couldn't open %s", fname);
	exit( 1 );
    }
    
    while (len--)
    {
	fprintf(fp, "%08lx", k[i]);
	i++;
    }
}


/*=======================================================
 * Generate a prime for the RSA key.  Not using DI's
 *  random #'s.  Using /dev/urandom instead.
 ======================================================*/
int genRsaPrime(DIGIT_T p[], unsigned int ndigits)
{
    FILE *fp;
    unsigned int i;
    
    for (i = 0; i < ndigits; i++)
	p[i] = spPseudoRand(0, MAX_DIGIT);

    fp = fopen( "/dev/urandom", "r" );
    if (fp == NULL)
    {
	printf("Couldn't open /dev/urandom");
	exit( 1 );
    }

    i = fread( p, sizeof(DIGIT_T), ndigits, fp );

#ifdef VERBOSE
    printf("p="); mpPrintNL(p, ndigits);
#endif


    /*	Make sure two highest and the low bits are set.
	Having the two highest bits set means the product
	(pq) will always have its highet bit set. 
    */
    p[ndigits - 1] |= HIBITMASK | (HIBITMASK >> 1);
    p[0] |= 0x1;

#ifdef VERBOSE
    printf("p'="); mpPrintNL(p, ndigits);
#endif 


    /* Check if prime */
    while (!mpIsPrime(p, ndigits, 10))
    {
	mpShortAdd(p, p, 2, ndigits);

#ifdef VERBOSE
	printf("p'="); mpPrintNL(p, ndigits);
#endif

	/* Check for overflow */
	if (!(p[ndigits - 1] & HIBITMASK))
	    return -1;	/* Failed to find a prime */
    }
    
    return 0;
}
