/* {{{ Preamble */

/* $Id: enforcer_lsm.c,v 1.68 2003/11/23 23:07:49 omen Exp $
 *
 * This is the Linux Enforcer Security Module
 *
 * Written and maintained by Omen Wild <Omen.Wild@Dartmouth.EDU>.
 * Feel free to email with questions or concerns.
 *
 * Copyright (C) 2003 Trustees of Dartmouth College
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/* }}} */

/* {{{ Includes */

#include "enforcer_lsm.h"

#include "enforcer_protos.h"

#include "bigint/undosig.h"

/* }}} */

/* {{{ defines/externs/globals */

extern long sys_sync(void);

static int secondary;			/* flag to keep track of how we were registered */

/* The hash database of files to protect */
static struct list_head db_entries[ENFORCER_HASH_SIZE];

/* Used to do a clean shutdown */
static struct rw_semaphore global_lock;

/* Original (dummy) security module. */
static struct security_operations *original_ops = NULL;

/* Minimal support for a secondary security module,
   just to allow the use of the dummy or capability modules. */
static struct security_operations *secondary_ops = NULL;

/* Command line parameters */
static char *enforcer_active = NULL;
static int enforcer_active_int = 1;
module_param_named(active, enforcer_active, charp, 0);
MODULE_PARM_DESC(active, "Pass 'no', 'off', 'false', or '0' to disable the enforcer module.");

#ifdef CONFIG_ENFORCER_DEBUG
int enforcer_debug_level = 0;
module_param_named(debug_level, enforcer_debug_level, int, 0);
MODULE_PARM_DESC(debug_level, "Set to > 0 to get debug info.");

static char *enforcer_check_signature = NULL;
static int enforcer_check_signature_int = 1;
module_param_named(check_signature, enforcer_check_signature, charp, 0);
MODULE_PARM_DESC(enforcer_check_signature, "Pass 'no', 'off', 'false', or '0' to not check database signature.");

#ifdef CONFIG_ENFORCER_TCPA
static char *enforcer_extend = NULL;
static int enforcer_extend_int = 1;
module_param_named(extend, enforcer_extend, charp, 0);
MODULE_PARM_DESC(extend, "Pass 'no', 'off', 'false', or '0' to not hash-and-extend with the public key.");
#endif /* CONFIG_ENFORCER_TCPA */
#endif /* CONFIG_ENFORCER_DEBUG */

/* }}} */

/* {{{ inline void enforcer_print_binary(const char *function, char *msg, u8 *binary, int len) */

#ifdef CONFIG_ENFORCER_DEBUG
inline void _enforcer_print_binary(const char *function, char *msg, u8 *binary, int len) {
   int i, j;
   char *buffer;
   int buffer_size = (len * 2) + 1 + (len / 2);
   /* Output: 616e:2041:4553:2073 */

   buffer = kmalloc(buffer_size, GFP_KERNEL);
   if(!buffer) {
	  printw (KERN_ERR, "unable to allocate %d bytes of memory.\n",
			 buffer_size);
	  return;
   }
   memset(buffer, 0, buffer_size);
   
   for (i = 0, j = 0; i < len && j < buffer_size; i++) {
	  if(i % 2 == 0 && i != 0)
		 buffer[j++] = ':';
	  
	  BUG_ON(j + 1 >= buffer_size);
	  buffer[j++] = enforcer_bin2hex(binary[i] >> 4);
	  buffer[j++] = enforcer_bin2hex(binary[i] & 0x0f);
   }

   printk(KERN_INFO MY_NAME ":%s: %s = %s\n",
		  function, msg, buffer);
   
   kfree(buffer);
}
#endif

/* }}} */

/* {{{ void enforcer_reverse_buffer(char *buffer, int len) */

void enforcer_reverse_buffer(char *buffer, int len) {	
   int i;
   
   for(i = 0; i < (len / 2); i++) {
	  int tmp = buffer[i];
	  buffer[i] = buffer[len - i - 1];
	  buffer[len - i - 1] = tmp;
   }
}

/* }}} */

/* {{{ static inline int __init enforcer_atoi(const char *name) */

/* Borrowed from drivers/video/modedb.c */
static inline int __init enforcer_atoi(const char *name)
{
   int val = 0;

   for (;; name++) {
	  switch (*name) {
	  case '0'...'9':
		 val = 10*val+(*name-'0');
		 break;
	  default:
		 return val;
	  }
   }
}

/* }}} */

/* {{{ inline char enforcer_bin2hex(int input) */

inline char enforcer_bin2hex(int input) {
   static const char bin2hex[] = { '0', '1', '2', '3',
								   '4', '5', '6', '7',
								   '8', '9', 'a', 'b',
								   'c', 'd', 'e', 'f' };

   if(input < 0x0 || input > 0xf) {
	  printw (KERN_WARNING, "ERROR, input value '%d' out of range (0..15)\n",
			  input);
	  
	  return '0';
   }

   return(bin2hex[input]);
}

/* }}} */

/* {{{ u8 enforcer_hex2bin(char input, char *input_txt) */

inline u8 enforcer_hex2bin(char input, char *input_txt) {
   char lcinput = __tolower(input);
   char output = 0;
	  
   if(lcinput >= '0' && lcinput <= '9') {
	  output = lcinput - '0';
   } else if(lcinput >= 'a' && lcinput <= 'f') {
	  output = lcinput - 'a' + 10;
   } else {
	  printd (1, " ERROR, unexpected input character '%c' in input string '%s'.\n",
			  input, input_txt);
   }

   printd(5, "input = %c output = %d\n", lcinput, output);
   return output;
}

/* }}} */

/* {{{ inline void enforcer_sha12str(char *sha1_txt, u8 *sha1) */

inline void enforcer_sha12str(char *sha1_txt, u8 *sha1) {
   int i;

   for (i = 0; i < SHA1_DIGEST_SIZE; ++i) {
	  sha1_txt[i * 2] = enforcer_bin2hex(sha1[i] >> 4);
	  sha1_txt[(i * 2) + 1] = enforcer_bin2hex(sha1[i] & 0x0f);
   }
}

/* }}} */

/* {{{ inline void enforcer_str2sha1(u8 *sha1, char *sha1_txt) */

inline void enforcer_str2sha1(u8 *sha1, char *sha1_txt) {
   int i;

   for (i = 0; i < SHA1_DIGEST_SIZE; i++) {
	  sha1[i] =
		 (enforcer_hex2bin(sha1_txt[i * 2], sha1_txt) << 4) |
		 (enforcer_hex2bin(sha1_txt[(i * 2) + 1], sha1_txt) & 0xf);
   }
}

/* }}} */

/* {{{ static void enforcer_dump_db_hash (void) */

#ifdef CONFIG_ENFORCER_DEBUG
static void enforcer_dump_db_hash (void) {
   int i;

   for(i = 0; i < ENFORCER_HASH_SIZE; i++) {
	  struct list_head *pos;
	  int count = 0;
	  struct list_head *list = &db_entries[i];

	  if(list_empty(list))
		 /* Skip non-existant entry */
		 continue;

	  list_for_each(pos, list) {
		 char sha1_txt[(SHA1_DIGEST_SIZE * 2) + 1];
		 struct db_entry *entry;

		 memset(sha1_txt, 0, sizeof(sha1_txt));
		 
		 count++;
		 
		 entry = list_entry(pos, struct db_entry, list);
		 enforcer_sha12str(sha1_txt, entry->sha1);

		 printd(3, "name = '%s' sha1 = %s action = %d\n",
				entry->file, sha1_txt, entry->action);
	  }

	  printd (0, "entries at slot %d = %d\n", i, count);
   }
   
}
#endif

/* }}} */

/* {{{ static inlined unsigned int enforcer_hash_calculate(char *filename) */

static inline unsigned int enforcer_hash_calculate(char *filename)
{
   unsigned int slot = 0;
   int len = strlen(filename);
   
   slot = full_name_hash(filename, len) % ENFORCER_HASH_SIZE;

   return slot;
}

/* }}} */

/* {{{ static int enforcer_do_readfile(char *filename, struct crypto_tfm *tfm, struct file *f) */

static int enforcer_do_readfile(char *filename, struct crypto_tfm *tfm, struct file *file) {
   char *buf;
   int len;
   int offset = 0;
   struct scatterlist sg[1];

   buf = (char *)kmalloc(PAGE_SIZE, GFP_KERNEL);
   if(!buf) {
	  printw (KERN_ALERT, "WARNING, unable to allocate %lu bytes of memory to check file '%s', file NOT checked.\n",
			  PAGE_SIZE, filename);
	  return -ENOMEM;
   }
   
   while((len = kernel_read(file, offset, buf, PAGE_SIZE)) > 0) {
	  offset += len;

	  sg[0].page = virt_to_page(buf);
	  sg[0].offset = offset_in_page(buf);
	  sg[0].length = len;
	  
	  crypto_digest_update(tfm, sg, 1);

	  printd(4, "read (%d) '%s'\n", len, buf);
   }

   kfree(buf);

   return len;
}

/* }}} */

/* {{{ static struct db_entry *enforcer_hash_lookup(char *filename, struct inode *inode) */

static struct db_entry *enforcer_hash_lookup(char *filename) {

   struct list_head *list;
   unsigned int slot;
   struct db_entry *entry;
   struct list_head *pos;

   slot = enforcer_hash_calculate(filename);
   
   list = &db_entries[slot];

   if(list_empty(list))
	  return NULL;

   list_for_each(pos, list) {
	  entry = list_entry(pos, struct db_entry, list);

	  if(!strcmp(entry->file, filename))
		 return entry;
   }

   return NULL;
}

/* }}} */

/* {{{ static inline void enforcer_insert_entry_hash (struct db_entry *entry) */

static inline void enforcer_insert_entry_hash (struct db_entry *entry) {

   struct list_head *list;
   unsigned int slot = enforcer_hash_calculate(entry->file);

   list = &db_entries[slot];

   list_add(&entry->list, list);
   
   return;
}

/* }}} */

/* {{{ static int enforcer_process_database_entry(...) */

static int enforcer_process_database_entry(char *file, int file_len,
										   char *sha1_txt, int sha1_txt_len,
										   char *action, int action_len)
{
   char *tok;
   struct db_entry *h;
   int error = 0;

   if(sha1_txt_len != SHA1_DIGEST_SIZE * 2) {
	  printw (KERN_ALERT, "WARNING, sha1 '%s' for file '%s' not the correct length (%d bytes).\n",
			  sha1_txt, file, SHA1_DIGEST_SIZE * 2);
	  return -EINVAL;
   }
   
   printd (3, "file = '%s' (%d) sha1 = %s action = %s\n",
		   file, file_len, sha1_txt, action);

   h = (struct db_entry *)kmalloc(sizeof(struct db_entry), GFP_KERNEL);
   if(!h) {
	  printw (KERN_ALERT, "WARNING, unable to allocate %d bytes of memory to check file '%s', file NOT checked.\n",
			  sizeof(struct db_entry), file);
	  return -ENOMEM;
   }
   memset(h, 0, sizeof(struct db_entry));

   h->file = (char *)kmalloc(file_len + 1, GFP_KERNEL);
   if(!h->file) {
	  printw (KERN_ALERT, "WARNING, unable to allocate %d bytes of memory to check file '%s', file NOT checked.\n",
			  file_len, file);
	  kfree(h);
	  return -ENOMEM;
   }
   h->file[file_len] = '\0';

   enforcer_str2sha1(h->sha1, sha1_txt);

   strncpy(h->file, file, file_len);

   while((tok = strsep(&action, ","))) {
	  if(!strnicmp("log", tok, 3)) {
		 h->action |= A_LOG;
	  } else if(!strnicmp("deny", tok, 4)) {
		 h->action |= A_DENY;
	  } else if(!strnicmp("panic", tok, 5)) {
		 h->action |= A_PANIC;
	  } else if(!strnicmp("tpm-lock", tok, 8)) {
		 h->action |= A_TPM_LOCK;
	  } else if(!strnicmp("call-helper", tok, 11)) {
		 h->action |= A_CALL_HELPER;
	  } else {
		 printw (KERN_WARNING, "unrecognized action '%s'.  Unable to load database.\n",
				 tok);
		 error = -EINVAL;
		 goto free_and_return;
	  }
   }

   if(h->action == 0) {
	  /* Config file entry didn't specify any action */
	  printw (KERN_WARNING, "config file entry for '%s' did not specify any action.  Unable to load database.\n",
			  file);
	  error = -EINVAL;
	  goto free_and_return;
   }

   enforcer_insert_entry_hash(h);
   return 0;
   
 free_and_return:
   kfree(h->file);
   kfree(h);
   
   return error;
}

/* }}} */

/* {{{ static int enforcer_calculate_sha1(u8 *sha1, struct file *file, char *filename) */

static int enforcer_calculate_sha1(u8 *sha1, struct file *file, char *filename) {
   struct crypto_tfm *tfm;
   int status;

   tfm = crypto_alloc_tfm("sha1", 0);
   if(tfm == NULL) {
	  printw(KERN_ERR, "failed to load transform for sha1.\n");
	  return -ENOMEM;
   }
   
   crypto_digest_init(tfm);

   if((status = enforcer_do_readfile(filename, tfm, file)) != 0) {
	  printw (KERN_WARNING, "unable to read contents of file '%s', file NOT checked.\n",
			 filename);
	  goto out;
   }

   crypto_digest_final(tfm, sha1);

 out:
   crypto_free_tfm(tfm);
   return status;
}

/* }}} */

/* {{{ static int enforcer_verify_db_sig(struct file *filp, char *filename) */

static int enforcer_verify_db_sig(struct file *filp, char *filename) {
   int error;
   u8 sha1_calculated[SHA1_DIGEST_SIZE];
   u8 sha1_expected[SHA1_DIGEST_SIZE];
   char sha1_expected_txt[(SHA1_DIGEST_SIZE * 2) + 1];

   u8 *signature = NULL;
   u8 *public_key = NULL;
   struct file *filp_signature = NULL, *filp_public_key = NULL;

   error = enforcer_calculate_sha1(sha1_calculated, filp, filename);
   
   if(error < 0) {
	  printw (KERN_ALERT, "WARNING, unable calculate a SHA1 of '%s'.\n",
			  filename);
	  return error;
   }

   signature = (u8 *)kmalloc(ENFORCER_SIGNATURE_SIZE, GFP_KERNEL);
   if(!signature) {
	  printw (KERN_ALERT, "WARNING, unable to allocate %d bytes of memory to check signature.\n",
			  ENFORCER_SIGNATURE_SIZE);
	  error = -ENOMEM;
	  return error;
   }

   public_key = (u8 *)kmalloc(ENFORCER_SIGNATURE_SIZE, GFP_KERNEL);
   if(!public_key) {
	  printw (KERN_ALERT, "WARNING, unable to allocate %d bytes of memory to check public key.\n",
			  ENFORCER_SIGNATURE_SIZE);
	  error = -ENOMEM;
	  goto out;
   }

   filp_signature = filp_open(ENFORCER_SIGNATURE_FILE, O_RDONLY, O_RDONLY);
   if (IS_ERR(filp_signature) || filp_signature == NULL) {
	  printw (KERN_ALERT, "WARNING, unable to open signature file '%s'.\n",
			  ENFORCER_SIGNATURE_FILE);
	  error = PTR_ERR(filp_signature);
	  goto out;
   }

   filp_public_key = filp_open(ENFORCER_PUBLIC_KEY_FILE, O_RDONLY, O_RDONLY);
   if (IS_ERR(filp_public_key) || filp_public_key == NULL) {
	  printw (KERN_ALERT, "WARNING, unable to open public key file '%s'.\n",
			  ENFORCER_PUBLIC_KEY_FILE);
	  error = PTR_ERR(filp_public_key);
	  goto out;
   }

   memset(signature, 0, ENFORCER_SIGNATURE_SIZE);
   memset(public_key, 0, ENFORCER_SIGNATURE_SIZE);

   error = kernel_read(filp_signature, 0, signature, ENFORCER_SIGNATURE_SIZE);
   if(error < 0) {
	  printw (KERN_ALERT, "WARNING, unable to read from signature file '%s'.\n",
			  ENFORCER_SIGNATURE_FILE);
	  goto out;
   }

   error = kernel_read(filp_public_key, 0, public_key, ENFORCER_SIGNATURE_SIZE);
   if(error < 0) {
	  printw (KERN_ALERT, "WARNING, unable to read from public key file '%s'.\n",
			  ENFORCER_PUBLIC_KEY_FILE);
	  goto out;
   }

#ifdef CONFIG_ENFORCER_TCPA
   {
	  int tpm_result_len;
	  u8 sha1_public_key[SHA1_DIGEST_SIZE];
	  u8 tpm_result[256];

	  memset(tpm_result, 0, 256);
	  
	  error = enforcer_calculate_sha1(sha1_public_key, filp_public_key, ENFORCER_PUBLIC_KEY_FILE);
	  if(error) {
		 printw (KERN_ALERT, "WARNING, unable calculate a SHA1 of '%s'.\n",
				 ENFORCER_PUBLIC_KEY_FILE);
		 goto out;
	  }

#ifdef CONFIG_ENFORCER_DEBUG
	  if(enforcer_extend_int != 0) {
#endif
		 error = enforcer_tcpa_hash_and_extend(sha1_public_key, ENFORCER_PUBLIC_KEY_PCR);
		 if(error != 0) {
			printw (KERN_ALERT, "error doing a TCPA hash and extend of the public key.\n");
			error = -EPERM;
			goto out;
		 }
#ifdef CONFIG_ENFORCER_DEBUG
	  } else {
		 printw (KERN_INFO, "hash-and-extend of public key disabled by request.\n");
	  }
#endif

	  error = enforcer_tcpa_unseal(ENFORCER_TCPA_BLOB_FILE,
								   ENFORCER_TCPA_PASSWORD_FILE,
								   tpm_result, sizeof(tpm_result),
								   &tpm_result_len);
	  
	  if(error != 0 || tpm_result_len <= 0) {
		 printw (KERN_ALERT,
				 "error doing a TCPA unseal, one or more PCRs may be invalid: TPM error = %d\n",
				 error);
		 error = -EPERM;
		 goto out;
	  }

	  printw (KERN_INFO,
			 "TPM Unseal operation successful, secret accessible.\n");
	  
#ifdef CONFIG_ENFORCER_DEBUG
	  printd(0, "enforcer_tcpa_useal result is %d bytes long.\n", tpm_result_len);
	  printd(0, "unsealed data: '%s'\n", tpm_result);
	  enforcer_print_binary("unsealed data", tpm_result, tpm_result_len);
#endif /* CONFIG_ENFORCER_DEBUG */
   }
#endif /* CONFIG_ENFORCER_TCPA */
   
   memset(sha1_expected_txt, 0, sizeof(sha1_expected_txt));
   undoSignature(sha1_expected_txt, public_key, signature);

   enforcer_str2sha1(sha1_expected, sha1_expected_txt);
   
   if(memcmp(sha1_calculated, sha1_expected, SHA1_DIGEST_SIZE) != 0) {
	  /* Error, signature incorrect */
	  printw (KERN_ALERT, "WARNING, database (%s) signature incorrect.\n",
			  filename);
	  error = -EINVAL;
	  goto out;
   }

   printw (KERN_INFO, "database (%s) signature verified.\n",
		   filename);
   
 out:
   if(signature)
	  kfree(signature);

   if(public_key)
	  kfree(public_key);
   
   if(filp_signature != NULL && !IS_ERR(filp_signature))
	  fput(filp_signature);

   if(filp_public_key != NULL && !IS_ERR(filp_public_key))
	  fput(filp_public_key);
   
   return error;
}

/* }}} */

/* {{{ static int enforcer_read_database(char *filename) */

/* Intially based on lids_init.c:lids_read_cap */

static int enforcer_read_database(char *filename) {
   struct file *filp;
   int len;
   int offset = 0;
   int error = 0;
   int count = 0;
   
   char *buffer;
   int buffer_len = PAGE_SIZE * 2;

   filp = filp_open(filename, O_RDONLY, O_RDONLY);
   if (IS_ERR(filp) || filp == NULL) {
	  printw (KERN_ALERT, "WARNING, unable to open config file '%s'\n",
			  filename);
	  return PTR_ERR(filp);
   }
   
#ifdef CONFIG_ENFORCER_DEBUG
   if(enforcer_check_signature_int != 0) {
#endif
	  error = enforcer_verify_db_sig(filp, filename);
	  if(error < 0) {
		 fput(filp);
		 return error;
	  }
#ifdef CONFIG_ENFORCER_DEBUG
   } else {
	  printw (KERN_INFO, "database signature checking disabled by request.\n");
   }
#endif	  

   buffer = (char *)kmalloc(buffer_len, GFP_KERNEL);
   if(!buffer) {
	  printw (KERN_ALERT, "WARNING, unable to allocate %d bytes of memory to check file '%s', file NOT checked.\n",
			  buffer_len, filename);
	  fput(filp);
	  return -ENOMEM;
   }

   while((len = kernel_read(filp, offset, buffer, buffer_len - 1)) > 0) {
	  char *c = buffer;
	  char *d_file, *d_sha1, *d_action;
	  int d_file_len, d_sha1_len, d_action_len;
	  error = 0;
	  
	  buffer[len] = '\0';		/* Ensure NULL terminated */
	  
   begin_line:
	  while(*c && (isspace(*c) || *c == '\n'))
		 /* Skip whitespace / empty lines */
		 c++;

	  if(*c && *c == '#') {
		 /* Skip comment line */
		 while(*c && *c != '\n')
			c++;

		 if(*c)
			c++;				/* Skip past \n */

		 /* Skipped comment line, now skip whitespace */
		 goto begin_line;
	  }

	  /* At this point, c points to the first non-whitespace character
		 on a non-blank, non-comment line */
	  /*
	   * Config file format:
	   * sha1_hash action1,action2,action3 file name until newline
	   */

	  /* Find end of sha1 */
	  d_sha1 = c;
	  while(*c && *c != ' ' && *c != '\n')
		 c++;

	  d_sha1_len = (unsigned long int)c - (unsigned long int)d_sha1;

	  if(*c && *c == ' ') {
		 *c = '\0';				/* Convert space to null */
		 c++;
	  } else if (!*c) {
		 goto contin;
	  } else {
		 printw (KERN_ALERT, "ERROR parsing config file '%s' between character #%d and #%d, expecting space at end of SHA1.\n",
				 filename, offset, offset + len);
		 error = -EINVAL;
		 goto close_and_exit;
	  }

	  
	  /* Find end of action */
	  d_action = c;
	  while(*c && *c != ' ' && *c != '\n')
		 c++;

	  d_action_len = (unsigned long int)c - (unsigned long int)d_action;

	  if(*c && *c == ' ') {
		 *c = '\0';				/* Convert space to null */
		 c++;
	  } else {
		 printw (KERN_ALERT, "ERROR parsing config file '%s' between character #%d and #%d, expecting space at end of action.\n",
				 filename, offset, offset + len);
		 error = -EINVAL;
		 goto close_and_exit;
	  }

	  
	  /* Find end of filename */
	  d_file = c;
	  while(*c && *c != '\n')
		 c++;

	  d_file_len = (unsigned long int)c - (unsigned long int)d_file;

	  if(*c && *c == '\n') {
		 *c = '\0';				/* Convert return to null */
		 c++;
	  } else {
		 printw (KERN_ALERT, "ERROR parsing config file '%s' between character #%d and #%d, expected '\\n'.\n",
				 filename, offset, offset + len);
		 error = -EINVAL;
		 goto close_and_exit;
	  }
	  
	  error = enforcer_process_database_entry(d_file, d_file_len,
											  d_sha1, d_sha1_len,
											  d_action, d_action_len);
	  if(error)
		 goto close_and_exit;
	  
	  count++;

   contin:
	  /* Process rest of file */
	  /* Bad form, really should continue processing already read buffer */
	  offset += (unsigned long int)c - (unsigned long int)buffer;
   }

 close_and_exit:
   fput(filp);					/* Close the file */
   kfree(buffer);

   if(error < 0)
	  return error;

   return count;
}

/* }}} */

/* {{{ static int enforcer_verify_sha1(struct db_entry *h, u8 *sha1) */

static int enforcer_verify_sha1(struct db_entry *h, u8 *sha1) {
   char sha1_txt_expected[(SHA1_DIGEST_SIZE * 2) + 1];
   char sha1_txt_calculated[(SHA1_DIGEST_SIZE * 2) + 1];
   int error = 0;
   
   if(!memcmp(sha1, h->sha1, SHA1_DIGEST_SIZE))
	  return 0;

   /* SHA1 calculated for file does not match what was stored in the
	  database */

   sha1_txt_expected[SHA1_DIGEST_SIZE * 2] = '\0';
   sha1_txt_calculated[SHA1_DIGEST_SIZE * 2] = '\0';

   enforcer_sha12str(sha1_txt_expected, h->sha1);
   enforcer_sha12str(sha1_txt_calculated, sha1);

   if(h->action & A_LOG) {
	  printw (KERN_ALERT, "WARNING, SHA1 value incorrect:\n"
			  MY_NAME ": File: %s\n"
			  MY_NAME ": Expected:   %s\n"
			  MY_NAME ": Calculated: %s\n"
			  MY_NAME ": this means the file has been modified since the database was built.  Your system may be compromised.\n",
			  h->file, sha1_txt_expected, sha1_txt_calculated);
   }
   
   if(h->action & A_TPM_LOCK) {
#ifdef CONFIG_ENFORCER_TCPA
	  if(enforcer_tcpa_lock() != 0)
		 printw (KERN_WARNING, "TPM lock failed, your secret may be exposed or the TPM hardware may be acting strangely.\n");

#else
      printw (KERN_WARNING, "TPM lock requested but TCPA hardware support not included.\n");
#endif
   }

   if(h->action & A_CALL_HELPER) {
	  int status;
	  char *envp[] = {
		 "HOME=/",
		 "TERM=linux",
		 "PATH=/usr/sbin:/sbin:/bin:/usr/bin",
		 NULL };
	  char *argv[] = {
		 ENFORCER_TCPA_COMMAND,
		 "force-stop",
		 NULL };

	  printd(0, "calling helper to unmount the loopback filesystem.\n");
	  status = call_usermodehelper(argv[0], argv, envp, 1);
	  if(status != 0) {
		 printw (KERN_EMERG, "call to helper '%s' returned %d, panicking.\n",
				ENFORCER_TCPA_COMMAND, status);
#ifdef ENFORCER_BUILTIN
		 sys_sync();
#else
#warning Enforcer built as a module, will NOT be able to sync the filesystem on a failed call to the helper.
#endif
		 panic(MY_NAME
			   ":%s: call to helper '%s' returned %d.\n",
			   __FUNCTION__, ENFORCER_TCPA_COMMAND, status);
	  }
   }

   if(h->action & A_PANIC) {
	  printw (KERN_EMERG, "panicking by request for file '%s'.\n",
			 h->file);
#ifdef ENFORCER_BUILTIN
		 sys_sync();
#else
#warning Enforcer built as a module, will NOT be able to sync the filesystem on a panic action.
#endif
	  panic(MY_NAME ":%s: panicking by request for file '%s'.\n",
			__FUNCTION__, h->file);
   }
	  
   if(h->action & A_DENY)
	  error = -EPERM;
   
   return error;
}

/* }}} */

/* {{{ static struct vfsmount *enforcer_find_vfsmount(struct inode *inode, struct list_head **start) */

/* Caller need to mntput() when it is done using the vfsmount */

static struct vfsmount *enforcer_find_vfsmount(struct inode *inode, struct list_head **start) {
   struct dentry *root = dget(inode->i_sb->s_root->d_sb->s_root);
   struct namespace *namespace = current->namespace;
   struct list_head *pos;
   struct vfsmount *mnt = NULL;

   down_read(&namespace->sem);

   if(*start == NULL)
	  *start = &namespace->list;

   /* Cribbed from list_for_each() */
	for (pos = (*start)->next, prefetch(pos->next); pos != (&namespace->list);
		 pos = pos->next, prefetch(pos->next)) {

	  mnt = list_entry(pos, struct vfsmount, mnt_list);
	  if (mnt->mnt_root == root) {
		 mntget(mnt);
		 *start = pos;
		 printd(3, "     found mount (0x%x)\n", (unsigned int)mnt);
		 break;
	  }

	  mnt = NULL;				/* Setup for next time through loop */
   }

   up_read(&namespace->sem);
   dput(root);

   return mnt;
}

/* }}} */

/* {{{ static void enforcer_hash_init(void) */

static void enforcer_hash_init(void) {
   int i;

   init_rwsem(&global_lock);

   for(i = 0; i < ENFORCER_HASH_SIZE; i++) {
	  struct list_head *list = &db_entries[i];
	  INIT_LIST_HEAD(list);
   }
}

/* }}} */

/* {{{ static void enforcer_hash_exit(void) */

static void enforcer_hash_exit(void) {
   int count = 0;
   int i;

   for(i = 0; i < ENFORCER_HASH_SIZE; i++) {
	  struct list_head *list = &db_entries[i];
	  struct list_head *pos, *n;

	  list_for_each_safe(pos, n, list) {
		 struct db_entry *entry = list_entry(pos, struct db_entry, list);
		 count++;
		 
		 list_del(pos);

		 if(entry->file)
			kfree(entry->file);

		 kfree(entry);
	  }
	  BUG_ON(!list_empty(list));
   }

   printd(1, "freed %d entries from db_entries.\n", count);
}

/* }}} */

/* {{{ static int enforcer_read_database_init(char *db)  */

static int enforcer_read_database_init(char *db) {
   int status;
#ifdef CONFIG_ENFORCER_DEBUG_TIME
   struct timeval tv1, tv2;
   do_gettimeofday(&tv1);
#endif   
   
   printd(1, "loading database.\n");

   status = enforcer_read_database(db);

   if(status > 0) {
	  /* > 0 is a count of how many entries loaded */
#ifdef CONFIG_ENFORCER_DEBUG_TIME
   {
	  unsigned long deltau, deltas;


	  do_gettimeofday(&tv2);

	  deltau = (tv2.tv_usec - tv1.tv_usec);
	  deltas = (tv2.tv_sec - tv1.tv_sec);

	  if(deltas > 0) {
		 deltau += deltas * 1000 * 1000;
	  }
	  
	  printw (KERN_INFO, "database loaded, %d entries, elapsed = %lu.%06lu sec (%lu us).\n",
			 status, deltau / 1000000, deltau % 1000000, deltau);
   };
#else
   printw (KERN_INFO, "database loaded, %d entries.\n",
		   status);
#endif
	  status = 0;				/* return only -errno or 0 */
   } else if(status < 0) {
	  /* < 0 is an error loading the database */
	  enforcer_hash_exit();
	  
	  printw (KERN_WARNING, "FAILED to load database, not active (errno = %d).\n",
			  -status);
	  
#ifdef CONFIG_ENFORCER_TCPA
	  status = enforcer_tcpa_lock();
	  if(status != 0)
		 printw (KERN_WARNING, "TPM lock failed, your secret may be exposed or the TPM hardware may be acting strangely.\n");
#endif
   } else {
	  /* status == 0 */
	  /* 0 no errors, but loaded 0 entries */
	  printw (KERN_WARNING, "loaded database, but zero entries, not active.\n");
   }
   
   return status;
}

/* }}} */

/* {{{ static int enforcer_inode_permission (struct inode *inode, int mask, struct nameidata *nd) */

static int enforcer_inode_permission (struct inode *inode, int mask, struct nameidata *nd)
{

   char *filename = NULL, *filebuf = NULL;
   int error = 0;
   struct db_entry *entry;
   struct file *file = NULL;
   u8 sha1[SHA1_DIGEST_SIZE];
   int found = 0;
   
   struct vfsmount *vfs = NULL;
   struct list_head *vfs_pos = NULL;
   struct list_head *d_pos;
   
   struct dentry *d = NULL;
   int valid_dentry = 0;

#ifdef CONFIG_ENFORCER_DEBUG_TIME
   struct timeval tv1, tv2;

   do_gettimeofday(&tv1);
#endif

   /* ??? Check nd structure */
   
   list_for_each(d_pos, &inode->i_dentry) {
	  d = list_entry(d_pos, struct dentry, d_alias);
	  if(atomic_read(&d->d_count) > 0) {
		 valid_dentry = 1;
		 break;
	  }
   }

   if(valid_dentry == 0) {
	  printw (KERN_ERR, "unable to find a valid dentry for inode %lu '%s', denying access.\n",
			  d->d_inode->i_ino, d->d_name.name);
	  BUG();
	  return -EPERM;
   }

   if(!S_ISREG(inode->i_mode))
	  return 0;					/* Only want to check regular files */
   
   /* Is a regular rile */

#if 0
   printd(1, "nd = 0x%x, inode %lu '%s'\n",
		  (unsigned int)nd, d->d_inode->i_ino, d->d_name.name);
#endif
   
   /* TAKE LOCKS */
   down_read(&global_lock);
   down(&inode->i_sem);

   spin_lock(&dcache_lock);
   dget(d);
   spin_unlock(&dcache_lock);

   printd(1, "----- %20.20s inode: %5lu\n", d->d_name.name, inode->i_ino);

   filebuf = (char *)kmalloc(PAGE_SIZE, GFP_KERNEL);
   if(!filebuf) {
	  printw (KERN_ALERT, "WARNING, unable to allocate %lu bytes of memory to check file named '%s', inode #%lu.  No file check performed.\n",
			  PAGE_SIZE, d->d_name.name, inode->i_ino);
	  error = -ENOMEM;
	  goto out;
   }

   while((vfs = enforcer_find_vfsmount(inode, &vfs_pos)) != NULL) {
	  found++;					/* Keep a count of number of matching files */
	  file = NULL;
	  
	  filename = d_path(d, vfs, filebuf, PAGE_SIZE);

	  if(!filename || *filename != '/') {
		 printw (KERN_ALERT, "WARNING, unable to find root filename for inode #%lu.\n",
				 inode->i_ino);
		 error = -EINVAL;
		 goto contin;
	  }

	  if(filename == filebuf) {
		 printw (KERN_ALERT, "WARNING, path for inode #%lu was more than %lu bytes long, unable to check, automatic failure.\n",
				 inode->i_ino, PAGE_SIZE);
		 error = -EINVAL;
		 goto contin;
	  }
	  
	  entry = enforcer_hash_lookup(filename);
	  if(!entry) {
		 /* filename does not exist in list of watched files */
		 goto contin;
	  }

	  /* ??? Need locks? */
	  dget(d);
	  mntget(vfs);	  

	  file = dentry_open(d, vfs, O_RDONLY);
   
	  if (file == NULL || IS_ERR(file)) {
		 error = PTR_ERR(file);
		 printw (KERN_ALERT,
				 "WARNING, unable to open file '%s' for testing (errno %d).\n",
				 filename, -error);
		 goto contin;
	  }

	  error = enforcer_calculate_sha1(sha1, file, filename);
	  if(error)
		 goto contin;

	  error = enforcer_verify_sha1(entry, sha1);
	  
   contin:
	  mntput(vfs);

	  if(file != NULL && !IS_ERR(file))
		 fput(file);

	  if(error != 0)
		 break;
   }
   
   if(found == 0) {
	  error = -EINVAL;
	  printw (KERN_ALERT, "WARNING, unable to find vfsmount for inode #%lu, name '%s'\n",
			  inode->i_ino, d->d_name.name);
   }

#ifdef CONFIG_ENFORCER_DEBUG_TIME
   {
	  unsigned long deltau, deltas;
	  do_gettimeofday(&tv2);

	  deltau = (tv2.tv_usec - tv1.tv_usec);
	  deltas = (tv2.tv_sec - tv1.tv_sec);

	  if(deltas > 0) {
		 deltau += deltas * 1000 * 1000;
	  }
	  
	  printw (KERN_INFO, "%s (%lu bytes, %d alias) elapsed = %lu.%06lu sec (%lu us).\n",
			  filename, (unsigned long)inode->i_size, found,
			  deltau / 1000000, deltau % 1000000, deltau);
   }
#endif

 out:
   dput(d);
   
   /* release LOCKS */
   up(&inode->i_sem);
   up_read(&global_lock);

   if(filebuf) 
	  kfree(filebuf);
   
   return(error);
}

/* }}} */

/* {{{ static void enforcer_sb_post_mountroot (void) */

static void enforcer_sb_post_mountroot (void)
{
#ifdef ENFORCER_BUILTIN
   (void) enforcer_read_database_init(ENFORCER_DATABASE_NAME);
#endif
   
   return;
}

/* }}} */

/* {{{ enforcer_register, enforcer_unregister: control lsm stacking */

/* module stacking operations, copied from selinux */

int enforcer_register_security (const char *name, struct security_operations *ops)
{
	if (secondary_ops != original_ops) {
	   printw (KERN_INFO,
				"there is already a secondary security module registered.\n");
		return -EINVAL;
 	}
	
	secondary_ops = ops;

	printw (KERN_INFO, "registering secondary module %s\n",
	       name);

	return 0;
}

int enforcer_unregister_security (const char *name, struct security_operations *ops)
{
	if (ops != secondary_ops) {
	   printw (KERN_INFO,
				"someone tried to unregister a security module that is not registered.\n");
		return -EINVAL;
	}

	secondary_ops = original_ops;

	return 0;
}

/* }}} */

/* {{{ static struct security_operations enforcer_ops */

static struct security_operations enforcer_ops = {
   /* Normal capabilities */
   .ptrace = cap_ptrace,
   .capget = cap_capget,
   .capset_check = cap_capset_check,
   .capset_set = cap_capset_set,
   .capable = cap_capable,
   .netlink_send = cap_netlink_send,
   .netlink_recv = cap_netlink_recv,

   .bprm_compute_creds = cap_bprm_compute_creds,
   .bprm_set_security = cap_bprm_set_security,
   .bprm_secureexec = cap_bprm_secureexec,

   .task_post_setuid = cap_task_post_setuid,
   .task_reparent_to_init = cap_task_reparent_to_init,

   .syslog = cap_syslog,

   .vm_enough_memory = cap_vm_enough_memory,
   /* End normal capabilities */

   .register_security = enforcer_register_security,
   .unregister_security = enforcer_unregister_security,

   .inode_permission = enforcer_inode_permission,
   .sb_post_mountroot = enforcer_sb_post_mountroot,
};

/* }}} */

/* {{{ static void enforcer_security_fixup_ops (struct security_operations *ops) */

#if 0							/* Unused for now */
static void enforcer_security_fixup_ops (struct security_operations *ops) {

#define set_to_capability_if_null(function)								\
   do {																	\
	  if (ops->function == NULL) {										\
		 ops->function = cap_##function;								\
	  } else {															\
		 printw(KERN_WARNING, "Security operation for %s already set, not overriding with capabilities .\n", #function); \
	  }																	\
   } while (0)

   /* Begin normal capabilities */
   set_to_capability_if_null(ptrace);
   set_to_capability_if_null(capget);
   set_to_capability_if_null(capset_check);
   set_to_capability_if_null(capset_set);
   set_to_capability_if_null(capable);
   set_to_capability_if_null(netlink_send);
   set_to_capability_if_null(netlink_recv);

   set_to_capability_if_null(bprm_compute_creds);
   set_to_capability_if_null(bprm_set_security);
   set_to_capability_if_null(bprm_secureexec);

   set_to_capability_if_null(task_post_setuid);
   set_to_capability_if_null(task_reparent_to_init);

   set_to_capability_if_null(syslog);

   set_to_capability_if_null(vm_enough_memory);
   /* End normal capabilities */
} 
#endif

/* }}} */

/* {{{ module init, exit */

static int __init enforcer_init (void)
{

   if(enforcer_active && 
	  (!strcmp("no", enforcer_active) ||
	   !strcmp("off", enforcer_active) ||
	   !strncmp("disable", enforcer_active, 7))) { /* Catches 'disabled' too */
	  printw (KERN_INFO, "disabled by request, loaded but not active.\n");
	  enforcer_active_int = 0;
	  return 0;
   }


#ifdef CONFIG_ENFORCER_DEBUG
   if(enforcer_check_signature && 
	  (!strcmp("no", enforcer_check_signature) ||
	   !strcmp("off", enforcer_check_signature) ||
	   !strcmp("0", enforcer_check_signature) ||
	   !strncmp("disable", enforcer_check_signature, 7))) { /* Catches 'disabled' too */
	  enforcer_check_signature_int = 0;
   }

#ifdef CONFIG_ENFORCER_TCPA
   if(enforcer_extend &&
	  (!strcmp("no", enforcer_extend) ||
	   !strcmp("off", enforcer_extend) ||
	   !strcmp("0", enforcer_extend) ||
	   !strncmp("disable", enforcer_extend, 7))) { /* Catches 'disabled' too */
	  enforcer_extend_int = 0;
   }
#endif /* CONFIG_ENFORCER_TCPA */
#endif /* CONFIG_ENFORCER_DEBUG */

   enforcer_hash_init();

#ifdef CONFIG_ENFORCER_DEBUG
   if(enforcer_debug_level >= 2) 
	  enforcer_dump_db_hash();
#endif
	  
#ifdef ENFORCER_MODULE
   {
	  int status;
	  if((status = enforcer_read_database_init(ENFORCER_DATABASE_NAME)) < 0) {
		 return status;
	  }
   }
#endif
   
   /* register ourselves with the security framework */
   if (register_security (&enforcer_ops)) {
	  printw (KERN_INFO,
			  "failure registering with the kernel as the primary.\n");
	  /* try registering with primary module */
	  if (mod_reg_security (MY_NAME, &enforcer_ops)) {
		 printw (KERN_INFO,
				 "failure registering module with primary security module.\n");
		 return -EINVAL;
	  }
	  secondary = 1;
   }

   printw (KERN_INFO, "initialized, version: %s\n", VERSION);
   return 0;
}


/* ??? Comment locking stuff */
static void __exit enforcer_exit (void)
{

   down_write(&global_lock);

   if(!enforcer_active_int) {
	  printw (KERN_INFO, "inactive, but removing module.\n");
	  goto unlock_and_return;
   }
   
   /* remove ourselves from the security framework */
	if (secondary) {
		if (mod_unreg_security (MY_NAME, &enforcer_ops))
		   printw (KERN_INFO, "failure unregistering with primary module.\n");
		goto unlock_and_return;
	}
 
	if (unregister_security (&enforcer_ops)) {
	   printw (KERN_INFO, "failure unregistering with the kernel.\n");
		goto unlock_and_return;
	}

	enforcer_hash_exit();
	
	up_write(&global_lock);
	down_write(&global_lock);
	
	printw (KERN_INFO, "removed.\n");

 unlock_and_return:
	up_write(&global_lock);
	
	return;
}

module_init (enforcer_init);
module_exit (enforcer_exit);

MODULE_AUTHOR("Omen Wild <Omen.Wild@Dartmouth.EDU>");
MODULE_DESCRIPTION("LSM Enforcer Module " VERSION);
MODULE_LICENSE("GPL");

/* }}} */
