/* {{{ Preamble */

/* $Id: enforcer_tcpa.c,v 1.12 2003/11/18 21:16:08 omen Exp $
 *
 * This is the Linux Enforcer Security Module
 *
 * Written and maintained by Omen Wild <Omen.Wild@Dartmouth.EDU>.
 * Feel free to email with questions or concerns.
 *
 * Copyright (C) 2003 Trustees of Dartmouth College
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/* }}} */

#include <linux/fs.h>
#include <linux/random.h>

#include "enforcer_lsm.h"

#include "tcpa/buildbuff.h"
#include "tcpa/oiaposap.h"
#include "tcpa/hmac.h"
#include "tpm/linux/tpm.h"

#ifdef CONFIG_ENFORCER_DEBUG
extern int enforcer_debug_level;
#endif

/* {{{ int enforcer_tcpa_hash_and_extend(u8 *sha1, int pcr_number) */

int enforcer_tcpa_hash_and_extend(u8 *sha1, int pcr_number) {
	  
   char hash_and_extend_fmt[] = "00 C1 T 00 00 00 14 L %";
   u8 *buffer;
   int res;

   buffer = kmalloc(TCPA_MAX_BUFF_SIZE, GFP_KERNEL);
   if(!buffer)
	  return -ENOMEM;

   memset(buffer, 0, TCPA_MAX_BUFF_SIZE);
   
   res = buildbuff(hash_and_extend_fmt, buffer, pcr_number,
				   SHA1_DIGEST_SIZE, sha1);

   if(res <= 0) {
	  res = -EINVAL;
	  goto out;
   }

   res = TPM_Transmit(buffer);

 out:
   kfree(buffer);
   
   return res;
}

/* }}} */

/* {{{ int enforcer_tcpa_lock(void) */

/* This routine hashes random data into a PCR register so any
   subsequent unseals fail */

int enforcer_tcpa_lock(void) {
#ifdef CONFIG_ENFORCER_DEBUG
	  printd(0, "would lock the TPM, but in debug mode.\n");
	  return 0;
#else
   u8 data[SHA1_DIGEST_SIZE];

   get_random_bytes(data, SHA1_DIGEST_SIZE);

   return enforcer_tcpa_hash_and_extend(data, ENFORCER_PUBLIC_KEY_PCR);
#endif
}

/* }}} */

/* {{{ int enforcer_tcpa_unseal(...) */

/*
 * Arguments:
 *
 * blob_filename: the filename containing the binary blob the TPM was
 * sealed with.
 *
 * password_filename: the file that contains the two passwords to the
 * TPM, one per line, \n separated.
 *
 * result: a buffer that can be used to store the results of the
 * unseal.
 *
 * result_buffer_size: the number of bytes allocated for the result
 * buffer.
 *
 * tpm_result_len: the number of bytes actually read back from the
 * TPM.  
 *
 * If the result buffer was non-null and result_buffer_size is >=
 * tpm_result_len, then result will hold the data.  If you just want
 * to know if an unseal can be done and don't care about the data,
 * then pass result as NULL and result_buffer_size as 0.  If the
 * return status is 0 and tpm_result_len > 0, then the unseal was
 * successful.
 *
 */

int enforcer_tcpa_unseal(char *blob_filename,
						 char *password_filename,
						 u8 *result, int result_buffer_size,
						 int *tpm_result_len) {
   
   uint32_t res;
   u8 passhash1[SHA1_DIGEST_SIZE], passhash2[SHA1_DIGEST_SIZE];
   struct file *filp = NULL;
   loff_t size;
   char *buffer,*str, *buffer_current;

   buffer = kmalloc(TCPA_MAX_BUFF_SIZE, GFP_KERNEL);
   if(!buffer) {
	  res = -ENOMEM;
	  return res;
   }
   
   /* Read the password file */
   filp = filp_open(password_filename, O_RDONLY, O_RDONLY);
   if (IS_ERR(filp) || filp == NULL) {
	  printw (KERN_ALERT, "WARNING, unable to open TPM password file '%s'\n",
			  password_filename);
	  res = PTR_ERR(filp);
	  goto out;
   }


   /* Process the contents of the password file */
   res = kernel_read(filp, 0, buffer, TCPA_MAX_BUFF_SIZE);
   size = filp->f_dentry->d_inode->i_size;
   if(res != size) {
	  printw (KERN_ALERT, "ERROR, unable to read from file '%s'\n",
			  password_filename);
	  goto out;
   }

   buffer_current = buffer;
   str = strsep(&buffer_current, "\n");
   if(str == NULL) {
	  printw (KERN_ALERT, "ERROR, unable to parse password from file '%s'\n",
			  password_filename);
	  res = -EINVAL;
	  goto out;
   }
   sha1(str, strlen(str), passhash1);
   
   str = strsep(&buffer_current, "\n");
   if(str == NULL) {
	  printw (KERN_ALERT, "ERROR, unable to parse password from file '%s'\n",
			  password_filename);
	  res = -EINVAL;
	  goto out;
   }
   sha1(str, strlen(str), passhash2);

   /* Cleanup */
   memset(buffer, 0, TCPA_MAX_BUFF_SIZE); /* Clear passwords */
   fput(filp);

   /* Read the blob file */
   filp = filp_open(blob_filename, O_RDONLY, O_RDONLY);
   if (IS_ERR(filp) || filp == NULL) {
	  printw (KERN_ALERT, "WARNING, unable to open TPM blob file '%s'\n",
			  blob_filename);
	  res = PTR_ERR(filp);
	  goto out;
   }

   size = filp->f_dentry->d_inode->i_size;
   if(size > TCPA_MAX_BUFF_SIZE) {
	  kfree(buffer);
	  buffer = kmalloc(size, GFP_KERNEL);
	  if(!buffer) {
		 res = -ENOMEM;
		 goto out;
	  }
   } 

   res = kernel_read(filp, 0, buffer, size);
   if(res != size) {
	  res = -EPERM;
	  goto out;
   }

   res = TPM_Unseal(0x40000000,	/* ??? make a define, SRK */
					passhash1, passhash2,
					buffer, size,
					result, result_buffer_size,
					tpm_result_len);

 out:
   if(buffer)
	  kfree(buffer);

   if(filp != NULL && !IS_ERR(filp))
	  fput(filp);
		 
   return res;
}

/* }}} */
