/****************************************************************************/
/*                                                                          */
/*               TPM Specific HMAC routines                                 */
/*                                                                          */
/* This file is copyright 2003 IBM. See "License" for details               */
/****************************************************************************/

#include <linux/crypto.h>
#include <linux/blkdev.h>

#include "../enforcer_lsm.h"
extern int enforcer_debug_level;

#include "tcpa.h"
#include "hmac.h"


/****************************************************************************/
/*                                                                          */
/* Validate the HMAC in an AUTH1 response                                   */
/*                                                                          */
/* This function validates the Authorization Digest for all AUTH1           */
/* responses.                                                               */
/*                                                                          */
/* The arguments are...                                                     */
/*                                                                          */
/* buffer - a pointer to response buffer                                    */
/* command - the command code from the original request                     */
/* ononce - a pointer to a 20 byte array containing the oddNonce            */
/* key    - a pointer to the key used in the request HMAC                   */
/* keylen - the size of the key                                             */
/* followed by a variable length set of arguments, which must come in       */
/* pairs.                                                                   */
/* The first value in each pair is the length of the data in the second     */
/*   argument of the pair                                                   */
/* The second value in each pair is an offset in the buffer to the data     */
/*   to be included in the hash for the paramdigest                         */
/* There should NOT be pairs for the TCPA_RESULT or TCPA_COMMAND_CODE       */
/* The last pair must be followed by a pair containing 0,0                  */
/*                                                                          */
/****************************************************************************/
#if 0
int checkhmac1(unsigned char *buffer, uint32_t command,
               unsigned char *ononce, unsigned char *key, int keylen, ...)
{
    uint32_t bufsize;
    uint16_t tag;
    uint32_t ordinal;
    uint32_t result;
    unsigned char *enonce;
    unsigned char *continueflag;
    unsigned char *authdata;
    unsigned char testhmac[20];
    unsigned char paramdigest[20];
    SHA_CTX sha;
    unsigned int dlen;
    unsigned int dpos;
    va_list argp;

    bufsize = htonl(*(uint32_t *) (buffer + 2));
    tag = htons(*(uint16_t *) (buffer + 0));
    ordinal = ntohl(command);
    result = *(uint32_t *) (buffer + TCPA_RETURN_OFFSET);

    if (tag == TPM_TAG_RSP_COMMAND)
        return 0;
    if (tag != TPM_TAG_RSP_AUTH1_COMMAND)
        return -1;
    authdata = buffer + bufsize - TCPA_HASH_SIZE;
    continueflag = authdata - 1;
    enonce = continueflag - TCPA_NONCE_SIZE;
    SHA1_Init(&sha);
    SHA1_Update(&sha, &result, 4);
    SHA1_Update(&sha, &ordinal, 4);
    va_start(argp, keylen);
    for (;;) {
        dlen = (unsigned int) va_arg(argp, unsigned int);
        if (dlen == 0)
            break;
        dpos = (unsigned int) va_arg(argp, unsigned int);
        SHA1_Update(&sha, buffer + dpos, dlen);
    }
    va_end(argp);
    SHA1_Final(paramdigest, &sha);
    rawhmac(testhmac, key, keylen, TCPA_HASH_SIZE, paramdigest,
            TCPA_NONCE_SIZE, enonce,
            TCPA_NONCE_SIZE, ononce, 1, continueflag, 0, 0);
    if (memcmp(testhmac, authdata, dlen) != 0)
        return -1;
    return 0;
}
#endif

/****************************************************************************/
/*                                                                          */
/* Calculate HMAC value for an AUTH1 command                                */
/*                                                                          */
/* This function calculates the Authorization Digest for all OIAP           */
/* commands.                                                                */
/*                                                                          */
/* The arguments are...                                                     */
/*                                                                          */
/* digest - a pointer to a 20 byte array that will receive the result       */
/* key    - a pointer to the key to be used in the HMAC calculation         */
/* keylen - the size of the key in bytes                                    */
/* h1     - a pointer to a 20 byte array containing the evenNonce           */
/* h2     - a pointer to a 20 byte array containing the oddNonce            */
/* h3     - an unsigned character containing the continueAuthSession value  */
/* followed by a variable length set of arguments, which must come in       */
/* pairs.                                                                   */
/* The first value in each pair is the length of the data in the second     */
/*   argument of the pair                                                   */
/* The second value in each pair is a pointer to the data to be hashed      */
/*   into the paramdigest.                                                  */
/* The last pair must be followed by a pair containing 0,0                  */
/*                                                                          */
/****************************************************************************/
int authhmac(unsigned char *digest, unsigned char *key,
             unsigned int keylen, unsigned char *h1, unsigned char *h2,
             unsigned char h3, ...)
{
   unsigned char paramdigest[TCPA_HASH_SIZE];
   struct crypto_tfm *tfm;
   unsigned int dlen;
   unsigned char *data;
   unsigned char c;
   int status = 0;
   struct scatterlist sg[1];

   va_list argp;

   if (h1 == NULL || h2 == NULL)
	  return -1;

   tfm = crypto_alloc_tfm("sha1", 0);
   if(tfm == NULL) {
	  return -1;
   }
   
   crypto_digest_init(tfm);
	
   c = h3;
   va_start(argp, h3);

   for (;;) {
	  dlen = (unsigned int) va_arg(argp, unsigned int);
	  if (dlen == 0)
		 break;

	  data = (unsigned char *) va_arg(argp, int);
	  if (data == NULL) {
		 status = -1;
		 goto out;
	  }

	  sg[0].page = virt_to_page(data);
	  sg[0].offset = offset_in_page(data);
	  sg[0].length = dlen;
	  
	  crypto_digest_update(tfm, sg, 1);
   }
	
   va_end(argp);
   crypto_digest_final(tfm, paramdigest);

   rawhmac(digest, key, keylen,
		   TCPA_HASH_SIZE, paramdigest,
		   TCPA_NONCE_SIZE, h1,
		   TCPA_NONCE_SIZE, h2,
		   1, &c,
		   0, 0);

 out:
   crypto_free_tfm(tfm);
	   
   return status;
}

/****************************************************************************/
/*                                                                          */
/* Calculate Raw HMAC value                                                 */
/*                                                                          */
/* This function calculates an HMAC digest                                  */
/*                                                                          */
/* The arguments are...                                                     */
/*                                                                          */
/* digest - a pointer to a 20 byte array that will receive the result       */
/* key    - a pointer to the key to be used in the HMAC calculation         */
/* keylen - the size of the key in bytes                                    */
/* followed by a variable length set of arguments, which must come in       */
/* pairs.                                                                   */
/* The first value in each pair is the length of the data in the second     */
/*   argument of the pair                                                   */
/* The second value in each pair is a pointer to the data to be hashed      */
/*   into the paramdigest.                                                  */
/* The last pair must be followed by a pair containing 0,0                  */
/*                                                                          */
/****************************************************************************/
int rawhmac(unsigned char *digest, unsigned char *key,
            unsigned int keylen, ...)
{
   unsigned int dlen;
   unsigned char *data;
   va_list argp;
   int error = 0;
   struct scatterlist sg[1];
   struct crypto_tfm *tfm;

   tfm = crypto_alloc_tfm("sha1", 0);
   if(tfm == NULL) {
	  return -1;
   }
   crypto_hmac_init(tfm, key, &keylen);
   
   va_start(argp, keylen);
   for (;;) {
	  dlen = (unsigned int) va_arg(argp, unsigned int);
	  if (dlen == 0)
		 break;
	  
	  data = (unsigned char *) va_arg(argp, int);
	  if (data == NULL) {
		 error = -1;
		 goto out;
	  }

	  sg[0].page = virt_to_page(data);
	  sg[0].offset = offset_in_page(data);
	  sg[0].length = dlen;

	  crypto_hmac_update(tfm, sg, 1);
   }
   va_end(argp);

   crypto_hmac_final(tfm, key, &keylen, digest);

 out:
   crypto_free_tfm(tfm);
   return error;
}

/****************************************************************************/
/*                                                                          */
/* Perform a SHA1 hash on a single buffer                                   */
/*                                                                          */
/****************************************************************************/
void sha1(unsigned char *input, int len, unsigned char *output)
{
   struct crypto_tfm *tfm;
   struct scatterlist sg[1];

   sg[0].page = virt_to_page(input);
   sg[0].offset = offset_in_page(input);
   sg[0].length = len;

   tfm = crypto_alloc_tfm("sha1", 0);
   if(tfm == NULL) {
	  return;
   }
   
   crypto_digest_init(tfm);

   crypto_digest_update(tfm, sg, 1);
   crypto_digest_final(tfm, output);

   crypto_free_tfm(tfm);
}
