/****************************************************************************/
/*                                                                          */
/*                            Seal a Data File                              */
/*                                                                          */
/*  This file is copyright 2003 IBM. See "License" for details              */
/****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <tcpa.h>

#include <openssl/rsa.h>
#include <openssl/pem.h>
#include <openssl/evp.h>

#include <buildbuff.h>

#define TPM_TAG_RQU_AUTH2_COMMAND 0x00C3
#define TPM_ORD_ActivateIdentity 122

void endian_swap_buffer(unsigned char *data, int len) {
   /* Make the blob network order */

   int i;

   if(len <= 4)
	  return;
   
   for (i = 0; i < len; i += 4) {
	  uint32_t *t = (uint32_t *)&data[i];
	  printf("0x%08x -> ", *t);
	  *t = ntohl(*t);
	  printf("0x%08x\n", *t);
   }

}

#if 0
typedef struct symmetric_key {
   uint32_t algorithm;
   uint16_t encscheme;
   uint16_t size;
   unsigned char data[32];
} symmetric_key;

typedef struct asym_ca_contents {
   symmetric_key sessionKey;
   unsigned char idDigest[TCPA_HASH_SIZE];
} asym_ca_contents;
#endif

int main(int argc, char *argv[]) {
   /* Should be arguments */
   uint32_t idKey = 0xFFAD02;
   char *ownerPassword = "tnd0007";
   char *idPassword = "tnd0007";

   unsigned char ownerPasswordHash[TCPA_HASH_SIZE];
   unsigned char idPasswordHash[TCPA_HASH_SIZE];
   
   int ret;

   FILE *log = stdout;

   pubkeydata pubek;
   keydata identityKey;
   
   int blobSize;
   int blobSize_no;
   unsigned char blob[8196];
   
   FILE *fp;

   unsigned char pubkeyHash[TCPA_HASH_SIZE];
   unsigned char idDigest[TCPA_HASH_SIZE];

   RSA *rsa;

   /* TPM_TAG_RQU_AUTH2_COMMAND */
   unsigned char activateidentity[] = "00 C3 T l L l % L % o % L % o %";
   unsigned char build_key_fmt[] = "L S S L L L L @";

   unsigned char tcpadata[TCPA_MAX_BUFF_SIZE];
   unsigned char asym_ca_contents[TCPA_MAX_BUFF_SIZE];
   unsigned char pubkeybuff[TCPA_MAX_BUFF_SIZE];
   
   int asym_ca_contents_len;
   unsigned char padded[RSA_MODULUS_BYTE_SIZE];
   unsigned char tcpa_oaep_pad_str[] = { 'T', 'C', 'P', 'A' };

   
   uint32_t ordinal = htonl(TPM_ORD_ActivateIdentity);

   uint32_t idKeyAuthHandle;
   unsigned char nonceOdd[TCPA_HASH_SIZE];
   unsigned char idKeyLastNonceEven[TCPA_HASH_SIZE];

   unsigned char idKeyAuth[TCPA_HASH_SIZE];

   uint32_t authHandle;
   unsigned char authLastNonceEven[TCPA_HASH_SIZE];
   unsigned char ownerAuth[TCPA_HASH_SIZE];
   
   idKeyAuthHandle = 0xffffffff;
   authHandle = 0xffffffff;
				  
   
   /* Read in identityPubkeyDigest (written by our makeidentity) */
   fp = fopen("identity.key", "r");
   if(fp == NULL) {
	  fprintf(stderr, "Error opening 'identity.pub.sha1'\n");
	  exit(1);
   }
   fread(&identityKey, sizeof(identityKey), 1, fp);
   fclose(fp);


   /* Build the TCPA_PUBLIC_KEY data structure */
   ret = buildbuff(build_key_fmt, pubkeybuff,
				   identityKey.pub.algorithm,  /* L */
				   identityKey.pub.encscheme,  /* S */
				   identityKey.pub.sigscheme,  /* S */
				   
				   /* parmSize */
				   12,						  /* L */
				   
				   /* RSA_KEY_PARMS */
				   identityKey.pub.keybitlen,  /* L */
				   identityKey.pub.numprimes,  /* L */
				   0,						  /* L */
				   
				   identityKey.pub.keylength, identityKey.pub.modulus /* @ */
				   );
   
   if(ret <= 0) {
	  fprintf(stderr, "PubKeyExtract() returned %d\n", ret);
	  goto out;
   }

   
   /* Digest of identity public key */
   sha1(pubkeybuff, ret, idDigest);


   /* Build ASYM_CA_CONTENTS (supplied by privacy CA) */
   asym_ca_contents_len = buildbuff("L S S % %", asym_ca_contents, 
								   0x6,							  /* L */
								   0x0,							  /* S */
								   32,							  /* S */
								   32, "012345678901234567890123456789Aa", /* % */
								   TCPA_HASH_SIZE, idDigest		  /* % */
								   );

   if (asym_ca_contents_len <= 0) {
	  fprintf(stderr, "buildbuf() returned %d\n", asym_ca_contents_len);
	  goto out;
   }

   if (TPM_ReadPubek(&pubek, log))
	  exit(2);

   rsa = convpubkey(&pubek);
   
   OpenSSL_add_all_algorithms();

   /* Pad and then encrypt the ASYM_CA_CONTENTS using Pubek */
   ret = RSA_padding_add_PKCS1_OAEP(padded, RSA_MODULUS_BYTE_SIZE,
									asym_ca_contents, asym_ca_contents_len,
									tcpa_oaep_pad_str, sizeof(tcpa_oaep_pad_str));
   if (ret == 0) {
	  fprintf(stderr, "RSA_padding_add_PKCS1_OAEP failed\n"); 
	  ret = 1;
	  goto out;
   }

   blobSize = RSA_public_encrypt(RSA_MODULUS_BYTE_SIZE, padded,
								 blob, rsa, RSA_NO_PADDING);
   if(blobSize <= 0) {
	  fprintf(stderr, "RSA_public_encrypt() returned %d\n", blobSize);
	  exit(3);
   }
   blobSize_no = htonl(blobSize);


   sha1(ownerPassword, strlen(ownerPassword), ownerPasswordHash);
   sha1(idPassword, strlen(idPassword), idPasswordHash);

   /* Open OIAP Session #1 */
   ret = TPM_OIAP(&idKeyAuthHandle, idKeyLastNonceEven, log);
   if (ret) {
	  fprintf(stderr, "TPM_OIAP() #1 returned %d\n", ret);
	  goto out;
   }

   RAND_bytes(nonceOdd, TCPA_NONCE_SIZE);

   /* From seal.c: calculate KEY authorization HMAC value */
   ret = authhmac(idKeyAuth,
				  idPasswordHash, TCPA_HASH_SIZE,
				  idKeyLastNonceEven, nonceOdd,
				  0,

				  sizeof(ordinal), &ordinal,
				  sizeof(blobSize), &blobSize_no,
				  blobSize, blob,
				  
				  0, 0);
   if (ret < 0) {
	  fprintf(stderr, "authhmac() #1 returned %d\n", ret);
	  goto out;
   }
   

   /* Open OIAP Session #2 */
   ret = TPM_OIAP(&authHandle, authLastNonceEven, log);
   if (ret) {
	  fprintf(stderr, "TPM_OIAP() #2 returned %d\n", ret);
	  goto out;
   }

   /* From seal.c: calculate KEY authorization HMAC value */
   ret = authhmac(ownerAuth,
				  ownerPasswordHash, TCPA_HASH_SIZE,
				  authLastNonceEven, nonceOdd,
				  0,

				  sizeof(ordinal), &ordinal,
				  sizeof(blobSize), &blobSize_no,
				  blobSize, blob,
				  
				  0, 0);
   if (ret < 0) {
	  fprintf(stderr, "authhmac() #2 returned %d\n", ret);
	  goto out;
   }


   ret = buildbuff(activateidentity, tcpadata,
				   ordinal,						  /* l */
				   idKey,						  /* L */
				   blobSize_no,					  /* l */
				   blobSize, blob,				  /* % */

				   idKeyAuthHandle,				  /* L */
				   TCPA_HASH_SIZE, nonceOdd,	  /* % */
				   0,							  /* o */
				   TCPA_HASH_SIZE, idKeyAuth,	  /* % */

				   authHandle,					  /* L */
				   TCPA_HASH_SIZE, nonceOdd,	  /* % */
				   0,							  /* o */
				   TCPA_HASH_SIZE, ownerAuth	  /* % */
				   );

   if (ret <= 0) {
	  fprintf(stderr, "buildbuff(...) = %d\n", ret);
	  goto out;
   }

   ret = TPM_Transmit(tcpadata, log, "activateidentity");
   if (ret != 0) {
	  fprintf(stderr, "TPM_Transmit(tcpadata) = %d\n", ret);
	  goto out;
   }

   ret = ntohl(*(uint32_t *) & tcpadata[TCPA_RETURN_OFFSET]);
   if (ret != 0) {
	  fprintf(stderr, "TCPA_RESULT = %d\n", ret);
	  goto out;
   }

   {
	  int i;
	  uint32_t type = ntohl(*(uint32_t *) & tcpadata[TCPA_DATA_OFFSET]);
	  uint16_t scheme = ntohs(*(uint16_t *) & tcpadata[TCPA_DATA_OFFSET + 4]);
	  uint16_t size = ntohs(*(uint16_t *) & tcpadata[TCPA_DATA_OFFSET + 6]);

	  printf("TCPA_SYMMETRIC_KEY: type = %d scheme = %d, size = %d\nSession key: 0x",
			 type, scheme, size);

	  for(i = 0; i < size; i += 4) {
		 uint32_t t = ntohl(*(uint32_t *) & tcpadata[TCPA_DATA_OFFSET + 8 + i]);
		 printf("%04x ", t);
	  }
	  printf("\n");
   }
   
 out:
   if(idKeyAuthHandle != 0xffffffff) {
	  TPM_Terminate_Handle(idKeyAuthHandle, log);
   }

   if(authHandle != 0xffffffff) {
	  TPM_Terminate_Handle(authHandle, log);
   }
	  
   exit(ret);
}

